@@ -1,5 +1,6 @@
 """networked spam-signature detection client"""
 
+import re
 import os
 import os.path
 import socket
@@ -8,6 +9,7 @@ import cStringIO
 import getopt
 import tempfile
 import mimetools
+import multifile
 import sha
 
 import pyzor
@@ -15,20 +17,18 @@ from pyzor import *
 
 __author__   = pyzor.__author__
 __version__  = pyzor.__version__
-__revision__ = "$Id: client.py,v 1.41 2002-09-08 20:37:15 ftobin Exp $"
+__revision__ = "$Id: client.py,v 1.48 2003-02-01 10:29:42 ftobin Exp $"
 
 randfile = '/dev/random'
 
 
 class Client(object):
     __slots__ = ['socket', 'output', 'accounts']
-    ttl = 4
     timeout = 5
     max_packet_size = 8192
-    
+
     def __init__(self, accounts):
         signal.signal(signal.SIGALRM, handle_timeout)
-        
         self.accounts = accounts
         self.output   = Output()
         self.build_socket()
@@ -42,7 +42,7 @@ class Client(object):
         msg = InfoRequest(digest)
         self.send(msg, address)
         return self.read_response(msg.get_thread())
-        
+
     def report(self, digest, spec, address):
         msg = ReportRequest(digest, spec)
         self.send(msg, address)
@@ -58,18 +58,13 @@ class Client(object):
         self.send(msg, address)
         return self.read_response(msg.get_thread())
 
-    def shutdown(self, address):
-        msg = ShutdownRequest()
-        self.send(msg, address)
-        return self.read_response(msg.get_thread())
-
     def build_socket(self):
         self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
 
     def send(self, msg, address):
         msg.init_for_sending()
         account = self.accounts[address]
-        
+
         mac_msg_str = str(MacEnvelope.wrap(account.username,
                                            account.keystuff.key,
                                            msg))
@@ -114,25 +109,34 @@ class Client(object):
 
 class ServerList(list):
     inform_url = 'http://pyzor.sourceforge.net/cgi-bin/inform-servers-0-3-x'
-    
+
     def read(self, serverfile):
         for line in serverfile:
             orig_line = line
             line = line.strip()
-            if line and not line.startswith('#'):
+            if line and not line.startswith('#') and \
+                re.match('[a-zA-Z0-9.-]+:[0-9]+', line):
                 self.append(pyzor.Address.from_str(line))
 
 
 
 class ExecCall(object):
     __slots__ = ['client', 'servers', 'output']
-    
+
     # hard-coded for the moment
     digest_spec = DataDigestSpec([(20, 3), (60, 3)])
 
     def run(self):
         debug = 0
-        (options, args) = getopt.getopt(sys.argv[1:], 'dh:', ['homedir='])
+        log = None
+        options = None
+
+        try:
+            (options, args) = getopt.getopt(sys.argv[1:], 'dh:',
+                                            ['homedir=', 'log', 'help'])
+        except getopt.GetoptError:
+            self.usage()
+
         if len(args) < 1:
            self.usage()
 
@@ -141,30 +145,37 @@ class ExecCall(object):
         for (o, v) in options:
             if o == '-d':
                 debug = 1
-            elif o == '-h':
+            elif o in ('-h', '--help'):
                self.usage()
             elif o == '--homedir':
                 specified_homedir = v
-        
-        self.output = Output(debug=debug)
+            elif o == '--log':
+                log = 1
 
+        self.output = Output(debug=debug)
         homedir = pyzor.get_homedir(specified_homedir)
 
+        if log:
+            sys.stderr = open(homedir + "/pyzor.log", 'a')
+            sys.stderr.write("\npyzor[" + repr (os.getpid()) + "]:\n")
+
         config = pyzor.Config(homedir)
         config.add_section('client')
 
-        defaults = {'ServersFile': 'servers',
+        defaults = {'ServersFile':        'servers',
                     'DiscoverServersURL': ServerList.inform_url,
-                    'AccountsFile' : 'accounts',
+                    'AccountsFile':       'accounts',
+                    'Timeout':            str(Client.timeout),
                     }
 
         for k, v in defaults.items():
             config.set('client', k, v)
-            
+
         config.read(os.path.join(homedir, 'config'))
-        
+
         servers_fn = config.get_filename('client', 'ServersFile')
-    
+        Client.timeout = config.getint('client', 'Timeout')
+
         if not os.path.exists(homedir):
             os.mkdir(homedir)
 
@@ -176,6 +187,12 @@ class ExecCall(object):
 
 
         self.servers  = self.get_servers(servers_fn)
+        if not len(self.servers):
+            sys.stderr.write("no valid servers found\n")
+            # Remove the servers file as it only contains invalid entries,
+            # probably a HTTP error message.
+            os.remove(servers_fn)
+            sys.exit(1)
         self.client = Client(self.get_accounts(config.get_filename('client',
                                                                    'AccountsFile')))
 
@@ -197,10 +214,13 @@ class ExecCall(object):
     def usage(self, s=None):
         if s is not None:
             sys.stderr.write("%s\n" % s)
-        sys.stderr.write("""usage: %s [-d] [--homedir dir] command [cmd_opts]
+        sys.stderr.write("""
+usage: %s [-d] [--homedir dir] command [cmd_opts]
 command is one of: check, report, discover, ping, digest, predigest,
-                   genkey, shutdown
+                   genkey
+
 Data is read on standard input (stdin).
+
 """
                          % sys.argv[0])
         sys.exit(2)
@@ -208,9 +228,9 @@ Data is read on standard input (stdin).
 
 
     def ping(self, args):
-        getopt.getopt(args[1:], '')
-
-        if len(args) > 1:
+        try:
+            getopt.getopt(args[1:], '')
+        except getopt.GetoptError:
             self.usage("%s does not take any non-option arguments" % args[0])
 
         runner = ClientRunner(self.client.ping)
@@ -219,32 +239,25 @@ Data is read on standard input (stdin).
             runner.run(server, (server,))
 
         return runner.all_ok
-        
-
-    def shutdown(self, args):
-        (opts, args2) = getopt.getopt(args[1:], '')
-
-        if len(args2) > 1:
-            self.usage("%s does not take any non-option arguments" % args[0])
-
-        runner = ClientRunner(self.client.shutdown)
-
-        for arg in args2:
-            server = Address.from_str(arg)
-            runner.run(server, (server,))
-                    
-        return runner.all_ok
 
 
     def info(self, args):
-        getopt.getopt(args[1:], '')
-        
-        if len(args) > 1:
+        try:
+            (options, args2) = getopt.getopt(args[1:], '', ['mbox'])
+        except getopt.GetoptError:
             self.usage("%s does not take any non-option arguments" % args[0])
 
+        do_mbox = 'msg'
+
+        for (o, v) in options:
+            if o == '--mbox':
+                do_mbox = 'mbox'
+
         runner = InfoClientRunner(self.client.info)
 
-        for digest in FileDigester(sys.stdin, self.digest_spec):
+        for digest in get_input_handler(sys.stdin, self.digest_spec, do_mbox):
+            if not digest:
+                continue
             for server in self.servers:
                 response = runner.run(server, (digest, server))
 
@@ -252,38 +265,49 @@ Data is read on standard input (stdin).
 
 
     def check(self, args):
-        getopt.getopt(args[1:], '')
-
-        if len(args) > 1:
+        try:
+            (options, args2) = getopt.getopt(args[1:], '', ['mbox'])
+        except getopt.GetoptError:
             self.usage("%s does not take any non-option arguments" % args[0])
 
+        do_mbox = 'msg'
+
+        for (o, v) in options:
+            if o == '--mbox':
+                do_mbox = 'mbox'
+
         runner = CheckClientRunner(self.client.check)
 
-        for digest in FileDigester(sys.stdin, self.digest_spec):
+        for digest in get_input_handler(sys.stdin, self.digest_spec, do_mbox):
+            if not digest:
+                continue
             for server in self.servers:
-                response = runner.run(server, (digest, server))
-                
+                runner.run(server, (digest, server))
+
         return (runner.found_hit and not runner.whitelisted)
 
 
     def report(self, args):
-        (options, args2) = getopt.getopt(args[1:], '', ['mbox'])
-        do_mbox = False
-
-        if len(args2) > 1:
+        try:
+           (options, args2) = getopt.getopt(args[1:], '', ['mbox'])
+        except getopt.GetoptError:
             self.usage("%s does not take any non-option arguments" % args[0])
 
+        do_mbox = 'msg'
+
         for (o, v) in options:
             if o == '--mbox':
-                do_mbox = True
-                
+                do_mbox = "mbox"
+
         all_ok = True
 
-        for digest in FileDigester(sys.stdin, self.digest_spec, do_mbox):
+        for digest in get_input_handler(sys.stdin, self.digest_spec, do_mbox):
+            if not digest:
+                continue
             if not self.send_digest(digest, self.digest_spec,
                                     self.client.report):
                 all_ok = False
-        
+
         return all_ok
 
 
@@ -297,59 +321,64 @@ Data is read on standard input (stdin).
 
         for server in self.servers:
             runner.run(server, (digest, spec, server))
-        
+
         return runner.all_ok
 
 
     def whitelist(self, args):
-        (options, args2) = getopt.getopt(args[1:], '', ['mbox'])
-
-        if len(args2) > 1:
+        try:
+            (options, args2) = getopt.getopt(args[1:], '', ['mbox'])
+        except getopt.GetoptError:
             self.usage("%s does not take any non-option arguments" % args[0])
 
-        do_mbox = False
+        do_mbox = "msg"
 
         for (o, v) in options:
             if o == '--mbox':
-                do_mbox = True
-                
+                do_mbox = "mbox"
+
         all_ok = True
 
-        for digest in FileDigester(sys.stdin, self.digest_spec, do_mbox):
+        for digest in get_input_handler(sys.stdin, self.digest_spec, do_mbox):
+            if not digest:
+                continue
             if not self.send_digest(digest, self.digest_spec,
                                     self.client.whitelist):
                 all_ok = False
-        
+
         return all_ok
 
 
     def digest(self, args):
-        (options, args2) = getopt.getopt(args[1:], '', ['mbox'])
-
-        if len(args2) > 1:
+        try:
+           (options, args2) = getopt.getopt(args[1:], '', ['mbox'])
+        except getopt.GetoptError:
             self.usage("%s does not take any non-option arguments" % args[0])
 
-
-        do_mbox = False
+        do_mbox = "msg"
 
         for (o, v) in options:
             if o == '--mbox':
-                do_mbox = True
-                
-        for digest in FileDigester(sys.stdin, self.digest_spec, do_mbox):
+                do_mbox = "mbox"
+
+        for digest in get_input_handler(sys.stdin, self.digest_spec, do_mbox):
+            if not digest:
+                continue
             sys.stdout.write("%s\n" % digest)
 
         return True
 
 
     def print_digested(self, args):
-        getopt.getopt(args[1:], '')
-
-        if len(args) > 1:
+        try:
+            getopt.getopt(args[1:], '')
+        except getopt.GetoptError:
             self.usage("%s does not take any non-option arguments" % args[0])
 
+        do_mbox = "msg"
+
         def loop():
-            for digest in FileDigester(sys.stdin, self.digest_spec):
+            for digest in get_input_handler(sys.stdin, self.digest_spec, do_mbox):
                 pass
 
         modglobal_apply(globals(), {'DataDigester': PrintingDataDigester},
@@ -358,9 +387,9 @@ Data is read on standard input (stdin).
         return True
 
     def genkey(self, args):
-        getopt.getopt(args[1:], '')
-
-        if len(args) > 1:
+        try:
+            getopt.getopt(args[1:], '')
+        except getopt.GetoptError:
             self.usage("%s does not take any non-option arguments" % args[0])
 
         import getpass
@@ -414,7 +443,6 @@ Data is read on standard input (stdin).
                   'report':    report,
                   'ping' :     ping,
                   'genkey':    genkey,
-                  'shutdown':  shutdown,
                   'info':      info,
                   'whitelist': whitelist,
                   'digest':    digest,
@@ -428,7 +456,7 @@ class DataDigester(object):
     """The major workhouse class"""
     __slots__ = ['_atomic', '_value', '_used_line', '_digest',
                  'seekable']
-    
+
     # minimum line length for it to be included as part
     # of the digest.  I forget the purpose, however.
     # Someone remind me so I can document it here.
@@ -465,8 +493,8 @@ class DataDigester(object):
         self.seekable   = seekable
 
         (fp, offsets) = self.get_line_offsets(fp)
-        
-        # did we get an empty file?
+
+        # did we get an empty (parsed output) file?
         if len(offsets) == 0:
             return
 
@@ -486,9 +514,12 @@ class DataDigester(object):
     def handle_atomic(self, fp):
         """we digest everything"""
         self._atomic = True
-        fp.seek(0)
-        for line in fp:
-            self.handle_line(line)
+        try:
+            fp.seek(0)
+            for line in fp:
+                self.handle_line(line)
+        except:
+            pass
 
 
     def handle_pieced(self, fp, spec, offsets):
@@ -519,7 +550,7 @@ class DataDigester(object):
             cur_offset = fp.tell()
             newfp = None
         else:
-            # we need a seekable file because to make
+            # we need a seekable file to make
             # line-based skipping around to be more efficient
             # than loading the whole thing into memory
             cur_offset = 0
@@ -527,7 +558,7 @@ class DataDigester(object):
 
 
         offsets = []
-        
+
         for line in fp:
             norm = self.normalize(line)
             should_handle = self.should_handle_line(norm)
@@ -583,9 +614,9 @@ class DataDigester(object):
     def normalize(self, s):
         repl = self.unwanted_txt_repl
         s2 = s
+        s2 = self.longstr_ptrn.sub(repl, s2)
         s2 = self.email_ptrn.sub(repl, s2)
         s2 = self.url_ptrn.sub(repl, s2)
-        s2 = self.longstr_ptrn.sub(repl, s2)
         s2 = self.html_tag_ptrn.sub(repl, s2)
         # make sure we do the whitespace last because some of
         # the previous patterns rely on whitespace
@@ -608,35 +639,41 @@ class PrintingDataDigester(DataDigester):
 
 
 
-class FileDigester(BasicIterator):
-    __slots__ = ['digester']
-
-    def __init__(self, fp, spec, mbox=False):
-        self.digester = iter(get_file_digester(fp, spec, mbox))
-        self.output = pyzor.Output()
-
-    def next(self):
-        digest = self.digester.next()
-        self.output.debug("calculated digest: %s" % digest)
-        return digest
-
-
-
-def get_file_digester(fp, spec, mbox, seekable=False):
+def get_input_handler(fp, spec, style='msg', seekable=False):
     """Return an object that can be iterated over
     to get all the digests from fp according to spec.
     mbox is a boolean"""
-    if mbox:
+    if style == 'msg':
+        return filter(lambda x: x is not None,
+                      (DataDigester(rfc822BodyCleaner(fp),
+                                    spec, seekable).get_digest(),)
+                      )
+
+    elif style =='mbox':
         return MailboxDigester(fp, spec)
 
-    return (DataDigester(rfc822BodyCleaner(fp),
-                         spec, seekable).get_digest(),)
+    elif style == 'digests':
+        return JustDigestsIterator(fp)
 
+    raise ValueError, "unknown input style"
+
+
+class JustDigestsIterator(BasicIterator):
+    __slots__ = ['fp']
+
+    def __init__(self, fp):
+        self.fp = fp
+
+    def next(self):
+        l = fp.readline()
+        if not l:
+            raise StopIteration
+        return l.rstrip()
 
 
 class MailboxDigester(BasicIterator):
     __slots__ = ['mbox', 'digest_spec', 'seekable']
-    
+
     def __init__(self, fp, digest_spec, seekable=False):
         import mailbox
         self.mbox        = mailbox.PortableUnixMailbox(fp,
@@ -645,7 +682,12 @@ class MailboxDigester(BasicIterator):
         self.seekable    = seekable
 
     def next(self):
-        next_msg = self.mbox.next()
+        try:
+            next_msg = self.mbox.next()
+        except IOError:
+            print "Error: Please feed mailbox files in on stdin, i.e."
+            print "    pyzor digest --mbox < my_mbox_file"
+            next_msg = None
         if next_msg is None:
             raise StopIteration
         return DataDigester(next_msg, self.digest_spec,
@@ -655,64 +697,110 @@ class MailboxDigester(BasicIterator):
 
 class rfc822BodyCleaner(BasicIterator):
     __slots__ = ['fp', 'multifile', 'curfile', 'type']
-    
+
     def __init__(self, fp):
         msg            = mimetools.Message(fp, seekable=0)
-        self.type      = msg.getmaintype()
+        # Default type is text.  See #1529694.
+        self.type      = msg.getmaintype() or "text"
         self.multifile = None
         self.curfile   = None
 
+        # Check if we got a mail or not. Set type to binary if there is no
+        # 'From:' header and type is text/plain with encoding 7bit.
+        # 7bit is passed through anyway so nobody cares.
+        if (not msg.has_key("From") and self.type == 'text' and
+            msg.subtype == 'plain' and msg.getencoding() == '7bit'):
+            self.type = 'binary'
+
+        if self.type is '':
+            self.type = 'text'
+
         if self.type == 'text':
             encoding = msg.getencoding()
-            if encoding == '7bit':
-                self.curfile = msg.fp
-            else:
+            self.curfile = msg.fp
+            if encoding != '7bit':
+                # fix bad encoding name
+                if encoding == '8bits':
+                    encoding = '8bit'
+
+                import binascii
                 self.curfile = tempfile.TemporaryFile()
-                mimetools.decode(msg.fp, self.curfile, encoding)
-                self.curfile.seek(0)
-                
+                try:
+                    mimetools.decode(msg.fp, self.curfile, encoding)
+                except binascii.Error, e:
+                    sys.stderr.write("%s: %s\n" % (e.__class__, e))
+                    self.curfile = cStringIO.StringIO()
+                except ValueError, e:
+                    #sys.stderr.write("%s: %s\n" % (e.__class__, e))
+                    self.curfile = msg.fp
+                except multifile.Error, e:
+                    #sys.stderr.write("%s: %s\n" % (e.__class__, e))
+                    self.curfile = msg.fp
+                try:
+                    self.curfile.seek(0)
+                except:
+                    # If we get an error we'll pass the message anyway.
+                    pass
+
         elif self.type == 'multipart':
-            import multifile
-            self.multifile = multifile.MultiFile(msg.fp, seekable=False)
-            self.multifile.push(msg.getparam('boundary'))
-            self.multifile.next()
-            self.curfile = self.__class__(self.multifile)
+            try:
+                self.multifile = multifile.MultiFile(msg.fp, seekable=False)
+                self.multifile.push(msg.getparam('boundary'))
+                self.multifile.next()
+                self.curfile = self.__class__(self.multifile)
+            except (TypeError, AttributeError, multifile.Error):
+                # ignore errors, pass msg as is
+                self.curfile = msg.fp
+                self.type = 'binary'
 
 
         if self.type == 'text' or self.type == 'multipart':
             assert self.curfile is not None
+        elif self.type == 'binary':
+            try:
+                fp.seek(0)
+            except:
+                pass
+            self.curfile = fp
         else:
             assert self.curfile is None
 
-        
+
     def readline(self):
         l = ''
-        if self.type in ('text', 'multipart'):
-            l = self.curfile.readline()
+        try:
+            if self.type in ('text', 'multipart', 'binary'):
+                l = self.curfile.readline()
 
-        if self.type == 'multipart' and not l and self.multifile.next():
-            self.curfile = self.__class__(self.multifile)
-            # recursion.  Could get messy if
-            # we get a bunch of empty multifile parts
-            l = self.readline()
+            if self.type == 'multipart' and not l and self.multifile.next():
+                self.curfile = self.__class__(self.multifile)
+                # recursion.  Could get messy if
+                # we get a bunch of empty multifile parts
+                l = self.readline()
+        except (TypeError, AttributeError, multifile.Error):
+            pass
         return l
 
 
     def next(self):
-        l = self.readline()
+        try:
+            l = self.readline()
+        except multifile.Error, e:
+            sys.stderr.write("%s: %s\n" % (e.__class__, e))
+            raise StopIteration
+
         if not l:
             raise StopIteration
-        return l        
-
+        return l
 
 
 class ClientRunner(object):
     __slots__ = ['routine', 'all_ok']
-    
+
     def __init__(self, routine):
         self.routine = routine
         self.setup()
-        
+
     def setup(self):
         self.all_ok = True
 
@@ -736,7 +824,7 @@ class ClientRunner(object):
             self.all_ok = False
         sys.stdout.write(message + str(response.head_tuple())
                          + '\n')
-    
+
 
 
 class CheckClientRunner(ClientRunner):
@@ -750,10 +838,10 @@ class CheckClientRunner(ClientRunner):
         self.found_hit   = False
         self.whitelisted = False
         super(CheckClientRunner, self).setup()
-    
+
     def handle_response(self, response, message):
         message += "%s\t" % str(response.head_tuple())
-        
+
         if response.is_ok():
             wl_count = int(response['WL-Count'])
             if wl_count > 0:
@@ -763,7 +851,7 @@ class CheckClientRunner(ClientRunner):
                 count = int(response['Count'])
                 if count > 0:
                     self.found_hit = True
-            
+
             message += "%d\t%d" % (count, wl_count)
             sys.stdout.write(message + '\n')
         else:
@@ -774,11 +862,11 @@ class CheckClientRunner(ClientRunner):
 class InfoClientRunner(ClientRunner):
     def handle_response(self, response, message):
         message += "%s\n" % str(response.head_tuple())
-        
+
         if response.is_ok():
             count = int(response['Count'])
             message += "\tCount: %d\n" % count
-            
+
             if count > 0:
                 for f in ('Entered', 'Updated', 'WL-Entered', 'WL-Updated'):
                     if response.has_key(f):
@@ -793,9 +881,9 @@ class InfoClientRunner(ClientRunner):
                         if f is 'WL-Entered':
                             message += ("\tWhiteList Count: %d\n"
                                         % int(response['WL-Count']))
-                        
+
                         message += ("\t%s: %s\n" % (f, stringed))
-            
+
             sys.stdout.write(message)
         else:
             sys.stderr.write(message)
@@ -833,13 +921,13 @@ class Keystuff(tuple):
         # be removed
         if self[1] is None:
             raise ValueError, "no key information"
-        
+
         for x in self:
             if not (isinstance(x, long) or x is None):
                 raise ValueError, "Keystuff must be long's or None's"
 
         # make sure we didn't get all None's
-        if not filter(lambda x: x is not None, self): 
+        if not filter(lambda x: x is not None, self):
             raise ValueError, "keystuff can't be all None's"
 
     def from_hexstr(self, s):
@@ -869,10 +957,10 @@ class Keystuff(tuple):
 class AccountsDict(dict):
     """Key is pyzor.Address, value is Account
     When getting, defaults to anonymous_account"""
-    
+
     anonymous_account = Account((pyzor.anonymous_user,
                                  Keystuff((None, 0L))))
-    
+
     def __setitem__(self, k, v):
         typecheck(k, pyzor.Address)
         typecheck(v, Account)
@@ -893,7 +981,7 @@ class AccountsFile(object):
     host : port ; username : keystuff
     """
     __slots__ = ['fp', 'lineno', 'output']
-    
+
     def __init__(self, fp):
         self.fp     = fp
         self.lineno = 0
@@ -906,7 +994,7 @@ class AccountsFile(object):
         while 1:
             orig_line = self.fp.readline()
             self.lineno += 1
-            
+
             if not orig_line:
                 raise StopIteration
             line = orig_line.strip()
@@ -914,7 +1002,7 @@ class AccountsFile(object):
                 continue
             fields = line.split(':')
             fields = map(lambda x: x.strip(), fields)
-            
+
             if len(fields) != 4:
                 self.output.warn("account file: invalid line %d: wrong number of parts"
                                  % self.lineno)
@@ -939,5 +1027,5 @@ def handle_timeout(signum, frame):
 
 
 def download(url, outfile):
-    import urllib
-    urllib.urlretrieve(url, outfile)
+    import urllib2
+    open(outfile, "wb").write(urllib2.urlopen(url).read())
