@@ -4,7 +4,6 @@ from __future__ import division
 
 import os
 import sys
-import signal
 import SocketServer
 import time
 import gdbm
@@ -17,7 +16,7 @@ from pyzor import *
 
 __author__   = pyzor.__author__
 __version__  = pyzor.__version__
-__revision__ = "$Id: server.py,v 1.25 2002-09-08 03:33:44 ftobin Exp $"
+__revision__ = "$Id: server.py,v 1.29 2002-10-09 00:45:45 ftobin Exp $"
 
 
 class AuthorizationError(pyzor.CommError):
@@ -343,41 +342,77 @@ class Record(object):
 
 
 
-class DBHandle(object):
-    __slots__ = ['output']
-    db_lock = threading.Lock()
-    max_age = 3600*24*30*4   # 3 months
-    db      = None
+class DBHandle(Singleton):
+    __slots__ = ['output', 'initialized']
+    db_lock   = threading.Lock()
+    max_age   = 3600*24*30*4   # 3 months
+    db        = None
+    sync_period = 60
+    reorganize_period = 3600*24  # 1 day
 
     def __init__(self):
         assert self.db is not None, "database was not initialized"
-        self.db_lock.acquire()
 
     def initialize(self, fn, mode):
         self.output = Output()
         self.db = gdbm.open(fn, mode)
+        self.start_reorganizing()
+        self.start_syncing()
     initialize = classmethod(initialize)
 
-    def __del__(self):
-        self.db.sync()
-        self.db_lock.release()
-
+    def apply_locking_method(self, method, varargs=(), kwargs={}):
+        # just so we don't carry around a mutable kwargs
+        if kwargs == {}:
+            kwargs = {}
+        self.output.debug("acquiring lock")
+        self.db_lock.acquire()
+        self.output.debug("acquired lock")
+        try:
+            result = apply(method, varargs, kwargs)
+        finally:
+            self.output.debug("releasing lock")
+            self.db_lock.release()
+            self.output.debug("released lock")
+        return result
+    apply_locking_method = classmethod(apply_locking_method)
+    
     def __getitem__(self, key):
+        return self.apply_locking_method(self._really_getitem, (key,))
+    
+    def _really_getitem(self, key):
         return self.db[key]
 
     def __setitem__(self, key, value):
+        self.apply_locking_method(self._really_setitem, (key, value))
+
+    def _really_setitem(self, key, value):
         self.db[key] = value
 
-    def __delitem__(self, key):
-        del self.db[key]
+    def start_syncing(self):
+        self.apply_locking_method(self._really_sync)
+        self.sync_timer = threading.Timer(self.sync_period,
+                                          self.start_syncing)
+        self.sync_timer.start()
+    start_syncing = classmethod(start_syncing)
 
-    def cleanup(self):
-        self.output.debug("cleaning up the database")
+    def _really_sync(self):
+        self.db.sync()
+    _really_sync = classmethod(_really_sync)
+
+    def start_reorganizing(self):
+        self.apply_locking_method(self._really_reorganize)
+        self.reorganize_timer = threading.Timer(self.reorganize_period,
+                                                self.start_reorganizing)
+        self.reorganize_timer.start()
+    start_reorganizing = classmethod(start_reorganizing)
+
+    def _really_reorganize(self):
+        self.output.debug("reorganizing the database")
         key = self.db.firstkey()
         breakpoint = time.time() - self.max_age
 
         while key is not None:
-            rec = Record.from_str(self[key])
+            rec = Record.from_str(self.db[key])
             delkey = None
             if rec.r_updated < breakpoint:
                 self.output.debug("deleting key %s" % key)
@@ -385,14 +420,11 @@ class DBHandle(object):
             key = self.db.nextkey(key)
             if delkey:
                 del self.db[delkey]
-        
         self.db.reorganize()
-
+    _really_reorganize = classmethod(_really_reorganize)
 
 
 class Server(SocketServer.ThreadingUDPServer, object):
-    ttl = 4
-    timeout = 3
     max_packet_size = 8192
     time_diff_allowance = 180
 
@@ -435,8 +467,6 @@ class RequestHandler(SocketServer.DatagramRequestHandler, object):
 
 
     def handle(self):
-        do_exit = False
-        
         try:
             self._really_handle()
         except UnsupportedVersionError, e:
@@ -447,14 +477,10 @@ class RequestHandler(SocketServer.DatagramRequestHandler, object):
             # We assume that KeyErrors are due to not
             # finding a key in the RFC822 message
             self.handle_error(400, "Bad request: %s" % e)
-        except TimeoutError, e:
-            self.handle_error(503, "Gateway timeout: %s" % e)
         except AuthorizationError, e:
             self.handle_error(401, "Unauthorized: %s" % e)
         except SignatureError, e:
             self.handle_error(401, "Unauthorized, Signature Error: %s" % e)
-        except SystemExit, e:
-            do_exit = True
         except Exception, e:
             self.handle_error(500, "Internal Server Error: %s" % e)
             traceback.print_exc()
@@ -470,11 +496,6 @@ class RequestHandler(SocketServer.DatagramRequestHandler, object):
         self.output.debug("sending: %s" % repr(msg_str))
         self.wfile.write(msg_str)
         
-        if do_exit:
-            db_hold = DBHandle()  # to keep the db consistent
-            self.finish()
-            os.kill(self.server.pid, signal.SIGQUIT)
-
 
     def _really_handle(self):
         """handle() without the exception handling"""
@@ -597,14 +618,9 @@ class RequestHandler(SocketServer.DatagramRequestHandler, object):
         self.out_msg['WL-Count'] = "%d" % wl_count
 
 
-    def handle_shutdown(self):
-        raise SystemExit
-
-
     dispatches = { 'check':     handle_check,
                    'report':    handle_report,
                    'ping':      None,
                    'info':      handle_info,
-                   'shutdown':  handle_shutdown,
                    'whitelist': handle_whitelist,
                    }
